/*
 * Tderef.c
 *
 * Various dereference functions for the translation table
 *
 * 
 * 9/24/93    Communication Optimizations have been introduced
 *            Still in testing and tuning stage for these optimizations,
 *            so the code has been short-circuited not to use the new
 *            adaptive functions. - Mustafa
 * 9/10/93    A fully replicated case has been covered - Mustafa
 * 9/8/93     Bugs are fixed, code is finalized - Mustafa
 * 8/10/93    Initial Revision - Mustafa, Raja
 *
 */

/*
 * The following will short-circuit the adaptive dereference functions 
 * which are still in testing and tuning stage. The correctness 
 * and/or performance is NOT guaranteed for this version! (MU)
 */
#ifdef ADAPTIVE
#undef ADAPTIVE
#endif

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include "parti.h"


#define PID_ZERO   0

#define NOPAGE   -1
#define DR_PROC   1
#define DR_OFFS  -1
#define DR_BOTH   0

void   derefComm(), derefCommPaged();
void   _regdereference(), _regderefoffset(), _regderefproc();
void   mregdereference(), regdereference();



/*----------------------------------------------------------------------*/
/*                         CODE STARTS HERE                             */
/*----------------------------------------------------------------------*/
#define SWAPINT(x,y)    {int t; t = (x); (x) = (y); (y) = t;}



/*
 * dereference
 *
 * Standart form of dereference giving both offsets and procs.
 *
 */
void _dereference(trans_table,index_array,local,proc,ndata,aflag)
     int *index_array,ndata;
     int *local,*proc;
     TTABLE *trans_table;
     int aflag;
{
  DSTABLE *table;
  
  if ( trans_table->type == REGULAR ) 
    {
      mregdereference( trans_table->regptr->dist_type, 
		     trans_table->regptr->data_size,
		     index_array, local, proc, ndata, DR_BOTH); 
    }
  else if (aflag == OFF)
    { 
      assert ( trans_table->type ==  DISTRIBUTED );
      table = (DSTABLE *)trans_table->tabptr ;
      DerefDistTable(table, index_array, local, proc, ndata);
    }
  else if (aflag == ON)
    {
      assert ( trans_table->type ==  DISTRIBUTED );
      table = (DSTABLE *)trans_table->tabptr ;
      DerefAdaptive(table, index_array, local, proc, ndata, DR_BOTH);
    }
  else
    panic("dereference: Nonexisting branch have been taken");
} 




/*
 * derefproc
 *
 * User level dereference proc function 
 *
 */
int derefproc(trans_table,index_array,proc,ndata,aflag)
     int *index_array,ndata;
     int *proc;
     TTABLE *trans_table;
     int aflag;
{
  DSTABLE *table;
  
  if ( trans_table->type == REGULAR ) 
    {
      mregdereference( trans_table->regptr->dist_type, 
		     trans_table->regptr->data_size,
		     index_array,NULL,proc,ndata,DR_PROC); 
    }
  else if (aflag == OFF)
    { 
      assert ( trans_table->type ==  DISTRIBUTED );
      table = (DSTABLE *)trans_table->tabptr ;
      _DerefProc(table, index_array, proc, ndata);
    }
  else if (aflag == ON)
    {
      assert ( trans_table->type ==  DISTRIBUTED );
      table = (DSTABLE *)trans_table->tabptr ;
      DerefAdaptive(table, index_array, NULL, proc, ndata, DR_PROC);
    }
  else
    panic("derefproc: Nonexisting branch have been taken");
} 





/*
 * derefoffset
 *
 * User level dereference offset function 
 *
 */
int derefoffset(trans_table,index_array,local,ndata,aflag)
     int *index_array,ndata;
     int *local;
     TTABLE *trans_table;
     int aflag;
{
  DSTABLE *table;
  
  if ( trans_table->type == REGULAR ) 
    {
      mregdereference( trans_table->regptr->dist_type, 
		     trans_table->regptr->data_size,
		     index_array,local,NULL,ndata,DR_OFFS); 
    }
  else if (aflag == OFF)
    { 
      assert ( trans_table->type ==  DISTRIBUTED );
      table = (DSTABLE *)trans_table->tabptr ;
      _DerefOffset(table, index_array, local, ndata);
    }
  else if (aflag == ON)
    {
      assert ( trans_table->type ==  DISTRIBUTED );
      table = (DSTABLE *)trans_table->tabptr ;
      DerefAdaptive(table, index_array, local, NULL, ndata, DR_OFFS);
    }
  else
    panic("derefoffset: Nonexisting branch have been taken");
} 





/*
 * _DerefProc
 *
 * Dereference the distributed translation table, yielding only proc assignments
 *
 */
int _DerefProc(Table,IndexArray,Proc,Ndata)
     DSTABLE *Table;
     int *IndexArray,Ndata;
     int *Proc;
{
  int count, *Flag, *proc, *local;
  
  if (Table->repFactor >= 1.0)
    return(derefProcReplicated(Table, IndexArray, Proc, Ndata));

  Flag   = (int *)malloc(3*sizeof(int)*Ndata);
  proc   = Flag + Ndata;
  local  = proc + Ndata;
  
  /* Local Dereference, ie. local part of the Table */
  count = derefLocal(Table, IndexArray, Proc, NULL, proc, local, Flag, Ndata, DR_PROC);

  /* Nonlocal Dereference */
  derefComm(Table, Proc, NULL, proc, local, Flag, count, DR_PROC);
  
  free(Flag); 
} 




/*
 * _DerefOffset
 *
 * Dereference the distributed translation table, yielding only offset assignments
 *
 */
int _DerefOffset(Table,IndexArray,Offset,Ndata)
     DSTABLE *Table;
     int *IndexArray,Ndata;
     int *Offset;
{
  int count, *Flag, *proc, *local;
  
  if (Table->repFactor >= 1.0)
    return(derefOffsetReplicated(Table, IndexArray, Offset, Ndata));
  
  Flag   = (int *)malloc(3*sizeof(int)*Ndata);
  proc   = Flag + Ndata;
  local  = proc + Ndata;
  
  /* Local Dereference, ie. local part of the Table */
  count = derefLocal(Table, IndexArray, NULL, Offset, proc, local, Flag, Ndata, DR_OFFS);

  /* Nonlocal Dereference */
  derefComm(Table, NULL, Offset, proc, local, Flag, count, DR_OFFS);
  
  free(Flag); 
} 





/*
 * DerefAdaptive
 *
 * Adaptively dereference translation table, bringing in the accessed 
 * pages as a whole as long as memory permits. 
 *
 */
int DerefAdaptive(table, indices, local,proc, nindex, mode)
     DSTABLEPTR  table;
     intPtr indices, proc,local;
     int nindex, mode;
{
  intPtr memStore;               /* base pointer to the space allocated */
  intPtr flg, prc, lcl,cbuffer,  /* pointers into memory space */
         pageList, hitCount;
  int count,totalPages, pageCnt; /* local counters */
  int newsize,size;              /* memory space counters */
  int ncount, sindex;
  
  /* handle special cases of RF=1.0 and RF=0.0 */
  if (table->repFactor >= 1.0)
    {
      switch (mode)
	{
	case DR_OFFS:
	  derefOffsetReplicated(table, indices, local, nindex);
	  break;
	case DR_PROC:
	  derefProcReplicated(table, indices, proc, nindex);
	  break;
	case DR_BOTH:
	  derefOffsetReplicated(table,indices,local,nindex);
	  derefProcReplicated(table,indices,proc,nindex);
	  break;
	default:
	  panic("DerefAdaptive: unknown mode");
	}
      return;
    }
#ifdef ADAPTIVE
  else if (table->repFactor <= 0.0)
#else
  else 
#endif
    {
      switch (mode)
	{
	case DR_OFFS:
	  _DerefOffset(table, indices, local, nindex);
	  break;
	case DR_PROC:
	  _DerefProc(table, indices, proc, nindex);
	  break;
	case DR_BOTH:
	  _DerefOffset(table,indices,local,nindex);
	  _DerefProc(table,indices,proc,nindex);
	  break;
	default:
	  panic("DerefAdaptive: unknown mode");
	}
      return;
    }
  
#ifdef ADAPTIVE
  /* Allocate all necessary storage space */
  totalPages  = (table->numPages*PARTI_numnodes+1);
  size = (3*nindex + 2*totalPages)*sizeof(int);
  flg = memStore = (intPtr) malloc(size);
  prc = flg + nindex;
  lcl = prc + nindex;
  hitCount = lcl + nindex;
  pageList = hitCount + totalPages;
  
  /* compulsory initializations */
  for (count=0; count<totalPages; count++)
    hitCount[count] = 0;
  
  /* Local Dereference, ie. local part of the Table */
  count = derefLocalAdaptive(table, indices, proc, local, nindex, mode,
			     flg, prc, lcl, hitCount);

  /* Adaptive page replacements */
  derefPages(table, hitCount, pageList, nindex);

  /* Eliminate accesses to the pages brought in recently */
  ncount = reDerefLocal(table, pageList, count, lcl, prc, flg, hitCount);

  
  /* Handle the accesses to the pages that are not local to our memory */
  derefCommPaged(table,proc,local,prc,lcl,flg,pageList,count,ncount,mode,nindex);

  /* clean up */
  free(memStore);
#endif
}





#ifdef ADAPTIVE
/*
 * derefLocalAdaptive
 *
 * Dereferences local part of the translation table, using all the pages 
 * stored in local memory & gathers the page access statistics for
 * adaptive part
 *
 */
int  derefLocalAdaptive(table, indices, proc, local, nData, mode,
			flg, prc, lcl, hitCount)
     DSTABLEPTR table;
     intPtr indices, proc, local, flg, prc, lcl, hitCount;
     int    nData,mode;
{
  register int pageNum,pageOffset,i,count;
  
  switch (mode)
    {
    case DR_BOTH: /* offs+proc */
      for(count=i=0;i<nData;i++)
	{
	  /* find the page and offset in the page for reference i */
	  pageNum    = indices[i] / table->pageSize;
	  pageOffset = indices[i] % table->pageSize;
	  hitCount[pageNum]++;
	  
	  if(table->pageTable[pageNum] != NULL)
	    {
	      proc[i]    = table->pageTable[pageNum]->Proc[pageOffset];
	      local[i]   = table->pageTable[pageNum]->Offset[pageOffset];
	    }
	  else
	    {
	      prc[count] = pageNum;
	      lcl[count] = pageOffset;
	      flg[count++] = i;
	    }
	}
      break;

    case DR_OFFS: /* offs only */
      for(count=i=0;i<nData;i++)
	{
	  /* find the page and offset in the page for reference i */
	  pageNum    = indices[i] / table->pageSize;
	  pageOffset = indices[i] % table->pageSize;
	  hitCount[pageNum]++;
	  
	  if(table->pageTable[pageNum] != NULL)
	    {
	      local[i]    = table->pageTable[pageNum]->Offset[pageOffset];
	    }
	  else
	    {
	      prc[count] = pageNum;
	      lcl[count] = pageOffset;
	      flg[count++] = i;
	    }
	}
      break;
      
    case DR_PROC: /* proc only */
      for(count=i=0;i<nData;i++)
	{
	  /* find the page and offset in the page for reference i */
	  pageNum    = indices[i] / table->pageSize;
	  pageOffset = indices[i] % table->pageSize;
	  hitCount[pageNum]++;
	  
	  if(table->pageTable[pageNum] != NULL)
	    {
	      proc[i]    = table->pageTable[pageNum]->Proc[pageOffset];
	    }
	  else
	    {
	      prc[count] = pageNum;
	      lcl[count] = pageOffset;
	      flg[count++] = i;
	    }
	}
      break;

    default:      /* error */
      panic("derefLocalAdaptive : error in processing");
    }
  
  return(count);
}




/*
 * reDerefLocal()
 *
 * reDereference the non-local pages after the adaptive step, hoping that
 * some of the dereference queries can be answered using the new local
 * pages still to be brought in.
 *
 */
int reDerefLocal(table,pageList,count,lcl, prc, flg, pgt) 
     DSTABLEPTR table;
     intPtr pageList, lcl, prc, flg, pgt;
     int    count;
{
  int index, page, offset;
  int pageCnt, tp;
  int i, cnt, t;

  pageCnt = pageList[0];
  tp = table->numPages*PARTI_numnodes;
  for (i=0; i<tp; i++)
    pgt[i] = 0;
  
  /* Mark pages to be brought in */
  for (i=1; i<=pageCnt; ++i)
    pgt[pageList[i]] = 1;

  /* rederef the old stuff to prevent duplicates */
  /* prc, lcl, flg are reorganized s.t. all paged stuff is at the end */
  for (cnt=count-1, i=0;  i<=cnt; )
    {
      if (pgt[prc[i]])
	{
	  SWAPINT(prc[i], prc[cnt]);
	  SWAPINT(flg[i], flg[cnt]);
	  SWAPINT(lcl[i], lcl[cnt]);
	  cnt--;
	}
      else
	{
	  lcl[i] = (prc[i]%table->numPages)*table->pageSize+lcl[i];
	  prc[i] = prc[i] / table->numPages;
	  i++;
	}
    }

  return(i);
}







/*
 * derefPages
 *
 * Use the page access statistics gathered in the adaptive local dereference to decide
 * which pages will be replaced, and carry out the page fetches. The current policy is 
 * that the pages with high hits are the winners.
 *
 */
int derefPages(table, hitCount, buffer, ni)
     DSTABLEPTR table; /* translation table to work on */  
     intPtr hitCount;  /* Hit statistics of pages from the previous deref */
     intPtr buffer;
     int ni;
{
  intPtr pageList;  /* List of pages to be brought in */
  intPtr temp;      /* scratch buffer again */
  int pageCnt;      /* no of pages to be brought in */
  int availPages;   /* number of mobile pages in the processor */
  int totalPages;   /* total number of pages in the translation table */
  int minCut;       /* min. cut off point for a page to be stay alive */
  int myBase;       /* the lowest numbered page statically bound to the processor */
  int limit;        /* exact number of pages bound to our processor */
  int myCeil, sweep;
  int i, intCompare(), lcnt;
  
  totalPages = table->numPages * PARTI_numnodes;
  myBase = PARTI_mynode * table->numPages;
  myCeil = myBase + table->numPages;
  minCut = 1;
  
  /* sweep over pages and collect the lucky pages to be brought in */
  pageList = buffer;
  for (pageCnt=i=0; i<totalPages && table->Replication>0; i++)
    {
      if ((hitCount[i]>=minCut) && (table->pageTable[i]==NULL))
	{
	  pageList[++pageCnt] = i;
	  table->Replication--;
	  pagesIn++;
	}
    }
  
  for (i=0; i<totalPages;i++)
    if ((hitCount[i]<minCut) && (table->pageTable[i]!=NULL) && (i<myBase) 
	&& (i>myCeil))
      {
	table->pageTable[i] = freePage(table->pageTable[i]);
	table->Replication++;
	pagesOut++;
      }
  
  pageList[0] = (pageCnt == 0) ? NOPAGE : pageCnt;
}



void derefCommPaged(table, proc, offset, 
		    prc,lcl,flg,
		    pageList, offproc, comm, mode, ni)
     DSTABLEPTR table;
     intPtr proc, offset, prc, lcl, flg, pageList;
     int  offproc, comm, mode,ni;
{
  PAGEPTR newPage;                   /* pointer to the new page */
  intPtr  iReq,wrk,cSum,pReq,cReq;   /* communication amount counters */
  int     pMax;
  int    **buckets, *bctrs, **flags; /* comm requests in bucketized version + counters */
  int     msgid1, msgid2, msgid3;    /* Async message ids */
  int     i, cnt, to, fromproc, sizercvd;
  intPtr  indexBuffer, obuf, pbuf, psnd, osnd;
  int     indexLen, offsLen, procLen, siza, sizb, osiz, psiz;

  /* memory allocation */
  iReq = (intPtr) malloc(6*PARTI_numnodes*sizeof(int));
  wrk  = iReq + PARTI_numnodes;
  cSum = wrk  + PARTI_numnodes;
  pReq = cSum + PARTI_numnodes;
  cReq = pReq + PARTI_numnodes;
  bctrs= cReq + PARTI_numnodes;
  buckets = (int **) malloc(2*PARTI_numnodes*sizeof(int*));
  flags   = buckets + PARTI_numnodes;
  
  /* initialize communication request counters */
  for (i=0; i<PARTI_numnodes; i++)
    pReq[i] = iReq[i] = 0;
  
  /* bucketize all my individual communication */
  for (i=0; i<comm; i++)
    iReq[prc[i]]++;

  /* bucketize all individual pages */
  for (i=1; i<=pageList[0]; i++)
    pReq[pageList[i]/table->numPages]++;

  /* copy into summation buffer */
  for (pMax=i=0; i<PARTI_numnodes; i++)
    {
      cReq[i] = cSum[i] = iReq[i] + pReq[i] + ((pReq[i]>0)||(iReq[i]>0));
      if (pReq[i]>pMax) pMax = pReq[i];
    }

  /* initialize outgoing buckets */
  for (cnt=i=0; i<PARTI_numnodes; i++)
    {
      buckets[i] = (int *) malloc(sizeof(int)*cReq[i]);
      buckets[i][0] = (pReq[i]) ? pReq[i] : NOPAGE;
      bctrs[i] = 1;
      flags[i] = (int *) malloc(sizeof(int)*cReq[i]);
    }

  /* sum the schedule communication amount over all processors */
  PARTI_gisum(cSum, PARTI_numnodes, wrk);
  PARTI_gihigh(&pMax, 1, &i);

  /* check to see if I need to participate in the rest */
  if ((cSum[PARTI_mynode] == 0) && (offproc == 0))
    return;

  /* msg passing buffer allocation */
  indexLen = ni;
  procLen = offsLen  = (pMax*table->pageSize+ni);
  indexBuffer = (intPtr)malloc((indexLen+2*procLen+2*offsLen)*sizeof(int));
  obuf = indexBuffer + indexLen;
  pbuf = obuf + offsLen;
  psnd = pbuf + procLen;
  osnd = psnd + procLen;
  indexLen *= sizeof(int);
  offsLen = procLen = procLen*sizeof(int);
  
  /* bucketize the requested pages */
  for (i=0; i<pageList[0]; )
    {
      to = pageList[++i]/table->numPages;
      buckets[to][bctrs[to]] = pageList[i];
      flags[to][bctrs[to]] = pageList[i];
      bctrs[to]++;
    }

  /* now, bucketize the individual offsets */
  for (i=0; i<comm; i++)
    {
      to = prc[i];
      buckets[to][bctrs[to]] = lcl[i];
      flags[to][bctrs[to]] = flg[i];
      bctrs[to]++;
    }

#ifndef TIMING
  if (cReq[PARTI_mynode])
    {
      printf("\n no need to comm w/myself me:%d ir:%d pr:%d cr:%d", PARTI_mynode, iReq[PARTI_mynode],
	     pReq[PARTI_mynode], cReq[PARTI_mynode]);
      fflush(stdout);
    }
#endif

  /* post the initial recv set */
  if (cSum[PARTI_mynode]>0)
    msgid1 = irecv(MSG_TTABLE, indexBuffer, indexLen);

  /* communicate requests and get the data immediately in msg exchange */
  for (i=1; (i<PARTI_numnodes) || (cSum[PARTI_mynode]>0); i++)
    {
      /* find a node with communication > 0 */
      to = (PARTI_mynode+i)%PARTI_numnodes;
      while((cReq[to]==0)&&(i<PARTI_numnodes))
	to = (PARTI_mynode+(++i))%PARTI_numnodes;	

      if (cReq[to]>0)
	{
	  msgid2 = irecv(MSG_PROC+to, pbuf, procLen);
	  msgid3 = irecv(MSG_OFFS+to, obuf, offsLen);

#ifndef TIMING
	  if (bctrs[to]*sizeof(int)>indexLen) 
	    {
	      printf("\n size overflow(3) at:%d snd:%d rsv:%d", PARTI_mynode, bctrs[to]*sizeof(int), indexLen);
	      fflush(stdout);
	      exit(0);
	    }
#endif
	  csend(MSG_TTABLE, buckets[to], bctrs[to]*sizeof(int), to, PID_ZERO);
	}

      while (cSum[PARTI_mynode]>0)
	{
	  if ((cReq[to]>0) && msgdone(msgid2)) 
	    break;
	  
	  if (msgdone(msgid1))
	    {
	      /* message housekeeping */
	      sizercvd = infocount() / sizeof(int);
	      fromproc = infonode();
	      cSum[PARTI_mynode] -= sizercvd;

	      /* construct the list of pages to be shipped off */
	      siza = _getPages(table->pageTable, indexBuffer, psnd, osnd, table->pageSize);

	      /* construct the list of individuals to be shipped off */
	      sizb = _getElems(table->Lproc, table->Loffset, psnd, osnd, 
			       indexBuffer, sizercvd, table->pageSize, mode);
	      
	      /* ship all info now */
	      psiz = (((mode==DR_BOTH) || (mode==DR_PROC)) ? sizb : siza) * sizeof(int);
	      osiz = (((mode==DR_BOTH) || (mode==DR_OFFS)) ? sizb : siza) * sizeof(int);

#ifndef TIMING
	      if (osiz > offsLen) panic("size error");
	      if (psiz > procLen) panic("size error2");
#endif
	      csend(MSG_PROC+PARTI_mynode, psnd, psiz, fromproc, PID_ZERO);
	      csend(MSG_OFFS+PARTI_mynode, osnd, osiz, fromproc, PID_ZERO);

	      /* if there is still stuff to be processed, post a new rcv */
	      if (cSum[PARTI_mynode]>0)
		  msgid1 = irecv(MSG_TTABLE, indexBuffer, indexLen);
	    }
	}

      if (cReq[to])
	{
	  /* wait for the other part */
	  msgwait(msgid3);
	  
	  /* form the newly acquired pages */
	  _putPages(table->pageTable, buckets[to], obuf, pbuf, table->pageSize);
	  
	  /* put the individual elements in their correct places */
	  _putElems(flags[to],obuf,pbuf,proc,offset,bctrs[to],table->pageSize,mode);
	  
	  /* I've finished with processor 'to' */
	  cReq[to] = 0;
	}
    }

#ifndef TIMING
  for (i=0; i<pageList[0]; i++)
    if (table->pageTable[pageList[i+1]] == NULL)
      printf("\n(@%d) page not exist : %d", PARTI_mynode, pageList[i+1]), fflush(stdout);
#endif

  _derefResidue(table, proc, offset, prc, lcl, flg, comm, offproc, mode);

  /* memory gargabe collection*/
  for (i=0; i<PARTI_numnodes; i++)
    free(flags[i]), free(buckets[i]);
  free(iReq); 
  free(buckets);
  free(indexBuffer);
}





/*
 * _derefResidue
 *
 * Dereference proc and/or offset values for the elements that are eliminated
 * for duplication reasons in function rederefLocal. 
 *
 * 9/24/93 * Mustafa
 */
int _derefResidue(table, proc, local, prc, lcl, flg, comm, offproc, mode)
     DSTABLEPTR table;
     intPtr  proc, local, prc, lcl, flg;
     int     comm, offproc, mode;
{
  int i;

  switch(mode)
    {
    case DR_BOTH:
      for (i=comm; i<offproc; i++)
	{
	  proc[flg[i]] = table->pageTable[prc[i]]->Proc[lcl[i]];
	  local[flg[i]] = table->pageTable[prc[i]]->Offset[lcl[i]];
	}
      break;

    case DR_PROC:
      for (i=comm; i<offproc; i++)
	{
	  proc[flg[i]] = table->pageTable[prc[i]]->Proc[lcl[i]];
	}
      break;

    case DR_OFFS:
      for (i=comm; i<offproc; i++)
	{
	  local[flg[i]] = table->pageTable[prc[i]]->Offset[lcl[i]];
	}
      break;

    default:
      panic("_derefResidue: invalid mode requested");
      break;
    }
}




/*
 * _getElems
 *
 * This functions performs a specialized 'gather' on the translation table
 * proc&offs arrays according to the dereference mode, forming send lists
 * for proc and/or offset.
 *
 * 9/24/93 * Mustafa
 */
int _getElems(myProc, myOffs, psnd, osnd, key, count, pagesize, mode)
     intPtr myProc, myOffs, psnd, osnd, key;
     int    count, pagesize,mode;
{
  int dst, src;
  
  dst = (key[0]==NOPAGE) ? 0 : key[0]*pagesize;
  src = (key[0]==NOPAGE) ? 1 : key[0]+1;

  switch(mode)
    {
    case DR_BOTH:
      for ( ; src < count; src++, dst++)
	{
	  psnd[dst] = myProc[key[src]];
	  osnd[dst] = myOffs[key[src]];
	}
      break;
    case DR_OFFS:
      for ( ; src < count; src++, dst++)
	{
	  osnd[dst] = myOffs[key[src]];
	}
      break;
    case DR_PROC:
      for ( ; src < count; src++, dst++)
	{
	  psnd[dst] = myProc[key[src]];
	}
      break;
    default:
      panic("_getElems: Unknown value for mode");
      break;
    }

  return(dst);
}





/*
 * _getPages
 *
 * Given a list of page numbers, this function expands the page table
 * entries into pbuf and obuf both of which form the contents of pages
 * given in the page list. 
 *
 * 9/24/93 * Mustafa
 */
int _getPages(pageTable, rcvbuf, pbuf, obuf, pagesize)
     PAGEPTR *pageTable;
     intPtr  rcvbuf, pbuf, obuf;
     int     pagesize;
{
  register int sndpgs, sndctr, page;
  register int i,j;
  
  sndpgs = rcvbuf[0];

  /* get the proc&offset values from pageTable into the output buffer */
  for (sndctr=i=0; i<sndpgs; i++)
    {
      page = rcvbuf[i+1];

#ifndef TIMING
      if (pageTable[page] == NULL)
	  panic("Nonexisting pagehas been requested");
#endif
      
      for (j=0; j<pagesize; j++,sndctr++)
	{
	  pbuf[sndctr] = pageTable[page]->Proc[j];
	  obuf[sndctr] = pageTable[page]->Offset[j];
	}
    }
  
  return(sndctr);
}



/*
 * _putPages
 *
 * Given the list of pages requested from a single processor, and the contents
 * of those pages in the expanded form, this function forms the local replicated
 * pages by copying the input buffer to compacted page structures.
 *
 * 9/24/93 * Mustafa
 */
_putPages(pageTable,key,obuf,pbuf,pgsize)
     PAGEPTR *pageTable;
     intPtr  key, obuf, pbuf;
{
  PAGEPTR newPage;
  int     i,j, rcvcnt;

  for (rcvcnt=0,i=1; i<=key[0]; i++)
    {
      newPage = (PAGEPTR) malloc(sizeof(PAGE));
      newPage->Offset = (intPtr) malloc(pgsize*sizeof(int));
      newPage->Proc   = (intPtr) malloc(pgsize*sizeof(int));
      pageTable[key[i]] = newPage;
      
      /* put the bulk in the correct slot of the page */
      for (j=0; j < pgsize; j++,rcvcnt++)
	{
	  newPage->Offset[j] = obuf[rcvcnt];
	  newPage->Proc[j] = pbuf[rcvcnt];
	}
    }
}



/* 
 * _putElems
 *
 * Given the proc and/or offset values as received from one processor,
 * it copies those values into their respective places in local and/or
 * proc. It is called after each message has been received from a corresponding
 * processor to put the individual elements in their correct places in the
 * correct dereference mode.
 *
 * 9/24/93 * Mustafa 
 */
int  _putElems(key, obuf, pbuf, proc, locl, cnt, pgsize, mode)
     intPtr key, obuf, pbuf, proc, locl;
     int    cnt,pgsize, mode;
{
  int src, dst;
  int i;

  src   = (key[0]==NOPAGE) ? 1 : key[0]+1;
  dst   = (key[0]==NOPAGE) ? 0 : key[0]*pgsize;

  switch(mode)
    {
    case DR_BOTH:
      for ( ; src < cnt; src++, dst++)
	{
	  proc[key[src]] = pbuf[dst];
	  locl[key[src]] = obuf[dst];
	}
      break;
    case DR_OFFS:
      for ( ; src < cnt; src++, dst++)
	{
	  locl[key[src]] = obuf[dst];
	}
      break;
    case DR_PROC:
      for ( ; src < cnt; src++, dst++)
	{
	  proc[key[src]] = pbuf[dst];
	}
      break;
    default:
      panic("_putElems: Unknown mode value");
      break;
    }
}
#endif /* of Adaptive */




int derefProcReplicated(Table,IndexArray,Proc,Ndata)
     int *IndexArray,Ndata;
     int *Proc;
     DSTABLE *Table;
{
  int i;

  for (i=0; i<Ndata; i++)
    Proc[i] = Table->Lproc[IndexArray[i]];
}



int derefOffsetReplicated(Table,IndexArray,Proc,Ndata)
     int *IndexArray,Ndata;
     int *Proc;
     DSTABLE *Table;
{
  int i;

  for (i=0; i<Ndata; i++)
    Proc[i] = Table->Loffset[IndexArray[i]];
}




void mregdereference(dist,size,global_refs,local,proc,ndata,mode)
int dist;
int size;
int *global_refs;
int *local;
int *proc;
int ndata,mode;
{
  switch(mode)
    {
    case DR_BOTH:
      _regdereference(dist,size,global_refs,local,proc,ndata);
      break;
    case DR_OFFS:
      _regderefoffset(dist,size,global_refs,local,ndata);
      break;
    case DR_PROC:
      _regderefproc(dist,size,global_refs,proc,ndata);
      break;
    default:
      break;
    }
}



void regdereference(dist,size,global_refs,local,proc,ndata)
int dist;
int size;
int *global_refs;
int *local;
int *proc;
int ndata;
{
	mregdereference(dist, size, global_refs, local, proc, ndata, DR_BOTH);
}

void _regdereference(dist,size,global_refs,local,proc,ndata)
int dist;
int size;
int *global_refs;
int *local;
int *proc;
int ndata;
{
  int nprocs, myproc, blocksize1, blocksize2;
  int extra, limit, i;

  nprocs = PARTI_numnodes;
  myproc = PARTI_mynode;
  blocksize1 = size/nprocs;
  blocksize2 = blocksize1 + 1;
  extra = size % nprocs;
  limit = (extra * blocksize2) + blocksize1;

  switch (dist) 
    {
    case 1:
      if(!(extra)) 
	{
	  for(i=0;i< ndata;i++)
	    {
	      proc[i] = global_refs[i] / blocksize1;
	      local[i] = global_refs[i] % blocksize1;
	    }
	}
      else
	{
	  for(i=0;i< ndata;i++)
	    {
	      if(global_refs[i] < limit )
		{
		  proc[i] = global_refs[i] / blocksize2;
		  local[i] = global_refs[i] % blocksize2;
		}
	      else
		{
		  proc[i] = (global_refs[i] - limit) / blocksize1 + extra + 1;
		  local[i] = (global_refs[i] - limit) % blocksize1;
		}
	    }
	}
      break;

    case 2:
      for(i=0;i< ndata;i++)
	{
	  proc[i] = global_refs[i] % nprocs;
	  local[i] = global_refs[i] / nprocs;
	}
      break;

      default :
	fprintf(stderr,"User have to supply own regular derefernce!!\n");
    }
}




void _regderefoffset(dist,size,global_refs,local,ndata)
int dist;
int size;
int *global_refs;
int *local;
int ndata;
{
  int nprocs, myproc, blocksize1, blocksize2;
  int extra, limit, i;

  nprocs = PARTI_numnodes;
  myproc = PARTI_mynode;
  blocksize1 = size/nprocs;
  blocksize2 = blocksize1 + 1;
  extra = size % nprocs;
  limit = (extra * blocksize2) + blocksize1;

  switch (dist) 
    {
    case 1:
      if(!(extra)) 
	{
	  for(i=0;i< ndata;i++)
	    {
	      local[i] = global_refs[i] % blocksize1;
	    }
	}
      else
	{
	  for(i=0;i< ndata;i++)
	    {
	      if(global_refs[i] < limit )
		{
		  local[i] = global_refs[i] % blocksize2;
		}
	      else
		{
		  local[i] = (global_refs[i] - limit) % blocksize1;
		}
	    }
	}
      break;

    case 2:
      for(i=0;i< ndata;i++)
	{
	  local[i] = global_refs[i] / nprocs;
	}
      break;

      default :
	fprintf(stderr,"User have to supply own regular derefernce!!\n");
    }
}




void _regderefproc(dist,size,global_refs,proc,ndata)
int dist;
int size;
int *global_refs;
int *proc;
int ndata;
{
  int nprocs, myproc, blocksize1, blocksize2;
  int extra, limit, i;

  nprocs = PARTI_numnodes;
  myproc = PARTI_mynode;
  blocksize1 = size/nprocs;
  blocksize2 = blocksize1 + 1;
  extra = size % nprocs;
  limit = (extra * blocksize2) + blocksize1;

  switch (dist) 
    {
    case 1:
      if(!(extra)) 
	{
	  for(i=0;i< ndata;i++)
	    {
	      proc[i] = global_refs[i] / blocksize1;
	    }
	}
      else
	{
	  for(i=0;i< ndata;i++)
	    {
	      if(global_refs[i] < limit )
		{
		  proc[i] = global_refs[i] / blocksize2;
		}
	      else
		{
		  proc[i] = (global_refs[i] - limit) / blocksize1 + extra + 1;
		}
	    }
	}
      break;

    case 2:
      for(i=0;i< ndata;i++)
	{
	  proc[i] = global_refs[i] % nprocs;
	}
      break;

      default :
	fprintf(stderr,"User have to supply own regular derefernce!!\n");
    }
}



/*
 * DerefDistTable
 *
 * Dereference the distributed translation table.
 */
int DerefDistTable(Table,IndexArray,Local,Proc,Ndata)
     int *IndexArray,Ndata;
     int *Local,*Proc;
     DSTABLE *Table;
{
  int count, *Flag, *proc, *local;

  if (Table->repFactor >= 1.0)
    {
    derefProcReplicated(Table, IndexArray, Proc, Ndata);
    return(derefOffsetReplicated(Table, IndexArray, Local, Ndata));
    }
   
  Flag   = (int *)malloc(3*sizeof(int)*Ndata);
  proc   = Flag + Ndata;
  local  = proc + Ndata;

  /* Local Dereference, ie. local part of the Table */
  count = derefLocal(Table, IndexArray, Proc, Local, proc, local, Flag, Ndata, DR_BOTH);

  /* Nonlocal Dereference */
  derefComm(Table, Proc, Local, proc, local, Flag, count, DR_BOTH);

  free(Flag); 
} 




/*
 * derefComm
 *
 * Performs the irregular communication for the dereference, fetching only
 * the necessary elements from other processor's memories.
 *
 */
 void derefComm(table, proc, offset, localProc, localOffset, flag, count, mode)
     DSTABLEPTR table;
     intPtr proc, offset, localProc, localOffset, flag;
     int  count, mode;
{
  SCHED *sr;
  int   i,j;

  /* perform communication */
  sr = schedule(localOffset,localProc,count,1);

  if (mode <= 0)
    {
      igather(localOffset, table->Loffset, sr);
      for (i=0; i<count; i++)
	offset[flag[i]] = localOffset[i];
    }

  if (mode >= 0)
    {
      igather(localProc,table->Lproc,sr);
      for (i=0; i<count; i++)
	proc[flag[i]] = localProc[i];
    }

  free_sched(sr);
}




/*
 * derefLocal
 *
 * Dereferences local part of the translation table, using all the pages 
 * stored in local memory
 *
 */
int  derefLocal(table, indices, proc, offset, localProc, localOffset, flag, nData, mode)
     DSTABLEPTR table;
     intPtr indices, proc, offset, localProc, localOffset, flag;
     int    nData, mode;
{
  int pageNum, pageOffset;
  int count, i, lPage;

  /* full dereference */
  if (mode == 0)
    {
      for(count=i=0;i<nData;i++)
	{
	  /* find the page and offset in the page for reference i */
	  pageNum    = indices[i] / table->pageSize;
	  pageOffset = indices[i] % table->pageSize;
	  
	  if(table->pageTable[pageNum] != NULL)
	    {
	      proc[i]   = table->pageTable[pageNum]->Proc[pageOffset];
	      offset[i] = table->pageTable[pageNum]->Offset[pageOffset];
	    }
	  else
	    {
	      flag[count] = i;
	      localProc[count]   = pageNum / table->numPages;
	      localOffset[count++] = (pageNum%table->numPages)*table->pageSize+pageOffset;
	    }
	}
    }
  else if (mode > 0 )    /* proc dereference */
    {
      for(count=i=0;i<nData;i++)
	{
	  /* find the page and offset in the page for reference i */
	  pageNum    = indices[i] / table->pageSize;
	  pageOffset = indices[i] % table->pageSize;
	  
	  if(table->pageTable[pageNum] != NULL)
	    {
	      proc[i]   = table->pageTable[pageNum]->Proc[pageOffset];
	    }
	  else
	    {
	      flag[count] = i;
	      localProc[count]   = pageNum / table->numPages;
	      localOffset[count++] = (pageNum%table->numPages)*table->pageSize+pageOffset;
	    }
	}
    }
  else if (mode < 0)   /* local dereference */
    {
      for(count=i=0;i<nData;i++)
	{
	  /* find the page and offset in the page for reference i */
	  pageNum    = indices[i] / table->pageSize;
	  pageOffset = indices[i] % table->pageSize;
	  
	  if(table->pageTable[pageNum] != NULL)
	    {
	      offset[i] = table->pageTable[pageNum]->Offset[pageOffset];
	    }
	  else
	    {
	      flag[count] = i;
	      localProc[count]     = pageNum / table->numPages;
	      localOffset[count++] = (pageNum%table->numPages)*table->pageSize+pageOffset;
	    }
	}
    }

  return(count);
}




/*
 * getPage()
 *
 * Brings in the pages listed in pageList as a whole and binds them to their
 * local slots in the home processor.
 *
 */
getPage(table, pageList, pageCnt)
     DSTABLEPTR table;
     intPtr  pageList;
     int pageCnt;
{
  SCHED *sr;
  PAGEPTR newPage;
  intPtr Proc, Local;
  int i, j, totalPages, procNum, page;
  int cnt, offset, pageNum;
  
  /* buffer for data transport */
  Proc  = (intPtr) malloc(pageCnt*table->pageSize*sizeof(int));
  Local = (intPtr) malloc(pageCnt*table->pageSize*sizeof(int));

  for (cnt=i=0; i<pageCnt; i++)
    {
      page = pageList[i];
      procNum = page / table->numPages;
      offset  = (page % table->numPages)*table->pageSize;
      for (j=0; j<table->pageSize; j++)
	{
	  Proc[cnt]  = procNum;
	  Local[cnt++] = offset++;
	}
    }

  sr = schedule(Local, Proc, cnt, 1);
  igather(Local, table->Loffset, sr);
  igather(Proc, table->Lproc, sr);

  for (cnt=i=0; i<pageCnt; i++, cnt+=table->pageSize)
    {
      newPage = (PAGEPTR) malloc(sizeof(PAGE));
      newPage->Offset = (intPtr) malloc(table->pageSize*sizeof(int));
      newPage->Proc   = (intPtr) malloc(table->pageSize*sizeof(int));
      memcpy(newPage->Offset, &Local[cnt], table->pageSize*sizeof(int));
      memcpy(newPage->Proc, &Proc[cnt], table->pageSize*sizeof(int)); 
      table->pageTable[pageList[i]] = newPage;
    }

  free(Proc);
  free(Local);
  free_sched(sr);
}





int   panic(msg)
     char *msg;
{
  fprintf(stderr, "\n%s @ %d", msg, PARTI_mynode);
  fflush(stderr);
  exit(1);
}



int  message(msg)
     char *msg;
{
  fprintf(stdout,"\n%s (@%d)", msg, PARTI_mynode);
  fflush(stdout);
}
